local json = require("json")
local utils = require("utils")
local const = require("const")
local utils_device = { _version = "0.1.0" }

function utils_device.get_device_profile( pk, dn )

    gw_pk, _, _, _ = utils.get_gw_config()
    gw_tls = "{\"profile\":{\"productKey\":\"\",\"productName\":\"EdgeGateway\",\"manufacture\":\"Alibaba\",\"protocol\":\"fake_monitor\",\"description\":\"A fake device that represents a monitor.\"},\"properties\":[{\"identifier\":\"cpu_usage\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"float\",\"specs\":{\"unit\":\"%\",\"min\":0,\"max\":100,\"step\":0.01}},\"desc\":\"CPU usage.\"},{\"identifier\":\"cpu_core_number\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"int\",\"specs\":{\"unit\":\"\",\"min\":1,\"max\":128,\"step\":1}},\"desc\":\"CPU core number.\"},{\"identifier\":\"memory_usage\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"float\",\"specs\":{\"unit\":\"%\",\"min\":0,\"max\":100,\"step\":0.01}},\"desc\":\"Memory usage in percentage.\"},{\"identifier\":\"memory_total\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"string\"},\"desc\":\"Total memory in bytes.\"},{\"identifier\":\"memory_free\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"string\"},\"desc\":\"Free memory in bytes.\"},{\"identifier\":\"disk_usage\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"float\",\"specs\":{\"unit\":\"%\",\"min\":0,\"max\":100,\"step\":0.01}}},{\"identifier\":\"gateway_product_key\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"string\",\"specs\":{\"unit\":\"\"}}},{\"identifier\":\"gateway_version\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"string\"},\"desc\":\"gateway software version.\"},{\"identifier\":\"system_info\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"string\"},\"desc\":\"OS version.\"},{\"identifier\":\"gateway_ip\",\"accessMode\":\"r\",\"dataType\":{\"type\":\"string\",\"specs\":{\"unit\":\"\"}},\"desc\":\"gateway IP.\"}],\"services\":[{\"identifier\":\"RemoteTerminalSwitch\",\"name\":\"RemoteTerminalSwitch\",\"desc\":\"开启/关闭远程调试\",\"inputData\":[{\"identifier\":\"switcher\",\"name\":\"开关\",\"dataType\":{\"type\":\"int\",\"specs\":{\"length\":\"1\"}}}],\"outputData\":[],\"method\":\"thing.service.RemoteTerminalSwitch\"}],\"events\":[]}"

    if (pk == gw_pk) then
        msg_raw = gw_tls
    else
        _, msg_raw = utils.get_data(const.data_key.tsl.standard_tsl_prefix..pk)
    end
    
    if (msg_raw == nil) then
        return false, nil
    end

    local ok, msg = pcall(json.decode, msg_raw)
    if (ok) then
        return true, msg
    end

    return false, nil
end


function utils_device.get_factor_detail(request_args, profile_key)

    utils.check_login()
    utils.check_permission("DeviceManage", "GatewayManage")

    if (nil == request_args) then
        utils.error_exit(400, "invalid parameter")
    end

    pk = request_args["ProductKey"]
    dn = request_args["DeviceName"]
    factor_id = request_args["FactorId"]

    utils.check_pk(pk)
    utils.check_dn(dn)

    if (nil == factor_id) then
        utils.error_exit(400, "invalid parameter")
    end

    local ok, profile = utils_device.get_device_profile(pk, dn)
    if (nil == profile) then
        if (ok) then
            utils.log_info("profile no found: ", pk, dn)
            utils.error_exit(200)
        else
            utils.log_error("get profile failed: ", pk, dn)
            utils.error_exit(500)
        end
    end

    for _, property in ipairs(profile[profile_key]) do
        if (factor_id == property["identifier"]) then
            res = {}
            res["ProductKey"] = pk
            res["DeviceName"] = dn
            res["FactorId"] = factor_id
            res["FactorName"] = factor_id
            res["Detail"] = property
            res_raw = json.encode(res)
            utils.error_exit(200, res_raw)
            return
        end
    end

    utils.error_exit(400, "Invalid FactorId")
end


function get_wkn_obj_inf(pk, dn)
    cloudid = pk .. "_" .. dn
    wkn = const.cmp.wkn_prefix .. cloudid
    obj = const.cmp.obj_prefix .. cloudid
    inf = const.cmp.inf_prefix .. cloudid

    return wkn, obj, inf
end


function utils_device.get_property_value(pk, dn, ...)
    msg = {}
    msg["params"] = {}
    for _, key in ipairs(...) do
        table.insert(msg["params"], key)
    end

    msg_json = json.encode(msg)
    wkn, obj, inf = get_wkn_obj_inf(pk, dn)
    ok, msg, err = utils.dbus_call_edge(wkn, obj, inf, "callServices", "get", msg_json)
    if (ok) then
        ok, res = pcall(json.decode, msg[1])
        if (ok and 0 == res["code"]) then
            return res["params"]
        end
    end

    return nil
end


function utils_device.set_property_value(pk, dn, kv)
    msg = {}
    msg_json = "{\"params\":{"..kv.."}}" 
    wkn, obj, inf = get_wkn_obj_inf(pk, dn)
    ok, msg, err = utils.dbus_call_edge(wkn, obj, inf, "callServices", "set", msg_json)
    if (ok) then
        ok, res = pcall(json.decode, msg[1])
        if (ok and 0 == res["code"]) then
            return true
        end
    end

    return false
end


function utils_device.call_service(pk, dn, service_name, kv)
    msg = {}
    msg["params"] = kv
    msg_json = json.encode(msg)
    wkn, obj, inf = get_wkn_obj_inf(pk, dn)
    ok, msg, err = utils.dbus_call_edge(wkn, obj, inf, "callServices", service_name, msg_json)
    if (ok) then
        ok, res = pcall(json.decode, msg[1])
        if (ok and 0 == res["code"]) then
            return true, res["params"]
        else
            return false, msg[1]
        end
    end

    return false, nil
end

local function transfer_device_local_state(online_state)
    if ("True" == online_state) then
        return "Online"
    else
        return "Offline"
    end
end

local function transfer_device_state(activation_state, online_state)
    if ("notActivated" == activation_state) then
        return "Inactive" 
    elseif ("activated" == activation_state) then
        if ("True" == online_state) then
            return "Online"
        else
            return "Offline"
        end
    elseif ("activationFailed" == activation_state) then
        return "ActivationFailed"  
    else
        return "ActivationFailed"
    end
end

local function get_standard_tsl(pk)
    local _, msg = utils.get_data(const.data_key.tsl.standard_tsl_prefix..pk)
    
    if (msg == nil) then
        msg = "{}"
    end

    return msg
end

local function get_product_name_from_TSL(tsl)
    local args = json.decode(tsl) 
    if args then
        if args["profile"] then
            return args["profile"]["productName"]
        end
    else
        return ""
    end
end



local function get_dev_info(ret, msg_raw, err)
    local item = {}
    local num = 0
    if (ret and msg_raw) then
        local ok, msg = pcall(json.decode, msg_raw)
        if (ok and 0 == msg["code"]) then
            local local_state, cloud_state
            local dev_info = msg["params"]["devices"][1]
            if (dev_info) then
                local pk = dev_info["device"][1]
                local tsl = get_standard_tsl(pk)
                item["ProductName"] = get_product_name_from_TSL(tsl)
                item["ProductKey"] = pk
                item["DeviceName"] = dev_info["deviceName"]
                item["DriverName"] = dev_info["driverName"]
                item["DeviceLocalId"] = dev_info["device"][2]
                item["LocalState"] = transfer_device_local_state(dev_info["localOnline"])
                item["CloudState"] = transfer_device_state(dev_info["activation"], dev_info["cloudOnline"])
                item["DeviceTags"] = "" --utils.get_device_tags(item["ProductKey"], item["DeviceName"])
                item["LastLocalOnlineTime"] = tostring(dev_info["device"][3])
                item["LastCloudOnlineTime"] = tostring(dev_info["device"][3])
                item["LastLinkTime"] = item["LastLocalOnlineTime"]
                num = 1
            end
        end
    else
        utils.log_error("call getDeviceState failed: " .. err)
        utils.error_exit(500, "call getDeviceState failed")
    end
    return num, item
end

local function get_device_info_inner(pk_dn_list, is_local)
    local method = "getDeviceState"

    if (is_local == true) then
        method = "getDeviceState"
    else
        method = "getLocalDeviceState"
    end
    local ok, msg_ret, err = utils.dbus_call_edge(const.dimu.wkn,
                                const.dimu.obj,
                                const.dimu.inf,
                                method, --"getLocalDeviceState", --
                                '{"params": '..pk_dn_list..'}')
    local msg_raw = msg_ret[1]
    local item_table = {}
    if (ok and msg_raw) then
        local ok, msg = pcall(json.decode, msg_raw)
        if (ok and 0 == msg["code"]) then
            local local_state, cloud_state
            local tab_pk_pn = {}
            for _, dev_info in ipairs(msg.params.devices) do
                local item = {}
                local pk = dev_info["device"][1]
                if tab_pk_pn[ pk ] == nil then
                    local tsl = get_standard_tsl(pk)
                    tab_pk_pn[ pk ] = get_product_name_from_TSL(tsl)
                end
                item["ProductName"] = tab_pk_pn[ pk ]
                item["ProductKey"] = pk
                item["DeviceName"] = dev_info["deviceName"]
                item["DriverName"] = dev_info["driverName"]
                item["DeviceLocalId"] = dev_info["device"][2]
                item["LocalState"] = transfer_device_local_state(dev_info["localOnline"])
                item["CloudState"] = transfer_device_state(dev_info["activation"], dev_info["cloudOnline"])
                item["DeviceTags"] = "" --utils.get_device_tags(item["ProductKey"], item["DeviceName"])
                item["LastLocalOnlineTime"] = tostring(dev_info["device"][3])
                item["LastCloudOnlineTime"] = tostring(dev_info["device"][3])
                item["LastLinkTime"] = item["LastLocalOnlineTime"]
                table.insert(item_table, item)
            end
        end
    else
        utils.log_error("call getDeviceState failed: " .. err)
        utils.error_exit(500, "call getDeviceState failed")
    end
    return #item_table, item_table
end

function utils_device.get_device_info_by_list(pk_dn_list, is_local)
    return get_device_info_inner(pk_dn_list, true) 
end 

function utils_device.get_local_device_info_by_list(pk_dn_list, is_local)
    return get_device_info_inner(pk_dn_list, false) 
end 


function utils_device.get_device_info(pk, dn)
    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                const.dimu.obj,
                                const.dimu.inf,
                                "getDeviceState", --"getLocalDeviceState", --
                                '{"params": [["' .. pk .. '", "' .. dn .. '"]]}')
    local num, tab_dev =  get_dev_info(ok, msg_raw[1], err)
   
    if tab_dev then
        tab_dev["DeviceTags"] = utils.get_device_tags(pk, dn)
    end

    return num, tab_dev
end

local function spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys 
    if order then
        table.sort(keys, function(a,b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

function utils_device.get_dev_list_by_page( dev_list, page_size, cur_page, func_dev_detail , sort_method)
    local total = #dev_list
    local list = {}

    table.sort(dev_list, function (e1, e2)
        return e1 <= e2
        end)
    local tab_pk_dn_list = {}
    for _, dev in pairs(dev_list) do
        local res = utils.split_str(dev, "_")
        local pk = res[1]
        local dn = table.concat(res, "_", 2)
        local tab_tmp = {}
        if (pk and dn) then
            table.insert(tab_tmp, pk)
            table.insert(tab_tmp, dn)

            table.insert(tab_pk_dn_list, tab_tmp)
        end
    end
    local _, res_list = func_dev_detail(json.encode(tab_pk_dn_list))

    if page_size <= 0 then
        page_size = total
    end
    local min, max = utils.get_item_index(total, page_size, cur_page)
    local tab_tmp = {}
    for k,v in spairs(res_list, function(t,a,b)
                                if (sort_method == "TimeDesc") then
                                    return t[b]["LastCloudOnlineTime"] < t[a]["LastCloudOnlineTime"]
                                elseif (sort_method == nil or sort_method == "nil") then
                                    return false
                                else
                                    return t[b]["LastCloudOnlineTime"] > t[a]["LastCloudOnlineTime"]
                                end
                                end) 
    do
        table.insert(tab_tmp, v)
    end
    
    for k, v in pairs(tab_tmp) do
           if (k >= min) and (k <= max) then
               v["DeviceTags"] = utils.get_device_tags(v["ProductKey"], v["DeviceName"])
               table.insert(list, v)
        end
    end 
   return total, list
end


local function get_dev_list(method, ...)
    local dev_set = {}
    local new = true

    local Set = require("set")
    for _, selector in ipairs({...}) do
        -- {"code":0,"message":"success","params":{"devNum":2,"devList":["a1NkFqUzGCG_AirConditioningIvan","xsRKiFzhE8r_proxy"]}}
        local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                    const.dimu.obj,
                                    const.dimu.inf,
                                    method,
                                    selector)                                  
        if (ok) then
            local ok, msg = pcall(json.decode, msg_raw[1])
            if (ok and 0 == msg["code"]) then
                local newlist = msg["params"]["devList"]
                if (newlist) then
                    local new_set = Set:new(newlist)
                    if (new) then
                        dev_set = new_set
                        new = false
                    end
                    dev_set = dev_set * new_set
                end
            end
        else
            utils.log_error("get device list failed" .. msg_raw .. err)
            utils.error_exit(500, "get device list failed")
        end
    end

    return dev_set:tolist()
end

function utils_device.get_authed_device_list_by_json(json_param, tags, page_size, cur_page, sort_method)
    local list = {}
    local total = 0
    local dev_list = {}
    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                        const.dimu.obj,
                                        const.dimu.inf,
                                        "getDeviceList",
                                        json_param)
    if (ok and msg_raw) then
        local ok, msg = pcall(json.decode, msg_raw[1])
        if (ok and 0 == msg["code"]) then
            dev_list =  msg["params"]["devList"]
        else
            utils.error_exit(500, json.encode("failed to query: "..msg_raw[1]))
        end
    else
        utils.error_exit(500, json.encode("failed to query: "..err))
    end
    if(tags) then
        local params = {}
        params["params"] = {}
        params["params"]["operation"] = "and"
        params["params"]["tags"] = {}
        for _, item in ipairs(tags) do
            local res = utils.split_str(item, '=')
            params["params"]["tags"][res[1]] = res[2]
        end

        local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                        const.dimu.obj,
                                        const.dimu.inf,
                                        "getDeviceByTag",
                                        json.encode(params))
        if (ok and msg_raw) then
            local ok, msg = pcall(json.decode, msg_raw[1])
            if (ok and 0 == msg["code"]) then
                local Set = require("set")

                local dev_list = msg["params"]
                if (dev_list) then
                    local set_tag = Set:new(dev_list)
                    local set_authed = Set:new(dev_list)
                    local set_intersection = set_authed * set_tag
                    return utils_device.get_dev_list_by_page(set_intersection:tolist(), page_size, cur_page, utils_device.get_device_info_by_list, sort_method)
                end
            else
                utils.log_error("get device list failed" .. msg_raw .. err)
                utils.error_exit(500, "get device list failed")
            end
        end
    end

    total, list = utils_device.get_dev_list_by_page(dev_list, page_size, cur_page, utils_device.get_device_info_by_list, sort_method)
    return total, list
end


function utils_device.get_authed_device_list(pk, dn, driver_name, tags, page_size, cur_page, sort_method)
    local list = {}
    local total = 0

    if (pk) then
        if (dn) then
            local num, dev = utils_device.get_device_info(pk, dn)
            if (num) then
                table.insert(list, dev)
                total = 1
            end
        else
            local dev_list = get_dev_list('getDeviceList', 'productKey=' .. pk)
            total, list = utils_device.get_dev_list_by_page(dev_list, page_size, cur_page, utils_device.get_device_info_by_list, sort_method)
        end
    elseif(driver_name) then
        local dev_list = get_dev_list('getDeviceList', 'driverName=' .. driver_name)
        total, list = utils_device.get_dev_list_by_page(dev_list, page_size, cur_page, utils_device.get_device_info_by_list, sort_method)
    elseif(tags) then
        local params = {}
        params["params"] = {}
        params["params"]["operation"] = "and"
        params["params"]["tags"] = {}
        for _, item in ipairs(tags) do
            local res = utils.split_str(item, '=')
            params["params"]["tags"][res[1]] = res[2]
        end

        local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                        const.dimu.obj,
                                        const.dimu.inf,
                                        "getDeviceByTag",
                                        json.encode(params))
        if (ok and msg_raw) then
            local ok, msg = pcall(json.decode, msg_raw[1])
            if (ok and 0 == msg["code"]) then
                local Set = require("set")

                local dev_list = msg["params"]
                if (dev_list) then
                    local set_tag = Set:new(dev_list)
                    local set_authed = Set:new(get_dev_list('getDeviceList', ''))
                    local set_intersection = set_authed * set_tag
                    return utils_device.get_dev_list_by_page(set_intersection:tolist(), page_size, cur_page, utils_device.get_device_info_by_list, sort_method)
                end
            else
                utils.log_error("get device list failed" .. msg_raw .. err)
                utils.error_exit(500, "get device list failed")
            end
        end
    else
        local dev_list = get_dev_list('getDeviceList', '')
        total, list = utils_device.get_dev_list_by_page(dev_list, page_size, cur_page, utils_device.get_device_info_by_list, sort_method)
    end

    return total, list
end

function utils_device.get_unauth_device_info(pk, lid)
    local ok, msg_raw, err = utils.dbus_call_edge(const.dimu.wkn,
                                const.dimu.obj,
                                const.dimu.inf,
                                "getLocalDeviceState", --"getDeviceState",
                                '{"params": [["' .. pk .. '", "' .. lid .. '"]]}')
    local num, dev_info = get_dev_info(ok, msg_raw[1], err)
    if tab_dev then
        tab_dev["DeviceTags"] = utils.get_device_tags(tab_dev["ProductKey"], tab_dev["DeviceName"])
    end

    if (num) then
        return 1, { ProductKey = dev_info["ProductKey"],
                    DeviceLocalId = dev_info["DeviceLocalId"],
                    DriverName = dev_info["DriverName"],
                    LastLinkTime = dev_info["LastLocalOnlineTime"] }
    else
        return 0, nil
    end
end


function utils_device.get_unauth_device_list(driver_name, pk, lid, page_size, cur_page)
    local dev_list = {}
    if (pk) then
        if (lid) then
            local num, dev = utils_device.get_unauth_device_info(pk, lid)
            if (1 == num) then
                table.insert(dev_list, dev)
                return 1, dev_list
            else
                return 0, {}
            end
        else
            dev_list = get_dev_list('getLocalDeviceList', 'productKey=' .. pk, 'authorized=False')
        end
    elseif (driver_name) then
        dev_list = get_dev_list('getLocalDeviceList', 'driverName=' .. driver_name, 'authorized=False')
    else
        dev_list = get_dev_list('getLocalDeviceList', 'authorized=False')
    end
    return utils_device.get_dev_list_by_page(dev_list, page_size, cur_page, utils_device.get_local_device_info_by_list, nil)
end


function utils_device.set_3tuple(device_tuples)
    local ok, msg = utils.set_data(const.data_key.dimu.device_3tuple, device_tuples)
    if (ok) then
        return true
    end

    utils.log_error("set device tuple data failed" .. msg)
    utils.error_exit(500, "store device tuple failed")
end


function utils_device.get_3tuple()
    local ok, msg = utils.get_data(const.data_key.dimu.device_3tuple)
    if (ok) then
        return true, msg
    end

    utils.log_error("get device tuple data failed" .. msg)
    utils.error_exit(500, "get device tuple failed")
end


function utils_device.set_3tuple_quanlity( quanlity )
    local ok, msg = utils.set_data(const.data_key.dimu.device_3tuple_quanlity, tostring(quanlity))
    if (ok) then
        return true
    end

    utils.log_error("set device 3tuple quanlity failed" .. msg)
    utils.error_exit(500, "set device tuple data failed")
end


function utils_device.get_3tuple_quanlity( )
    local ok, msg = utils.get_data(const.data_key.dimu.device_3tuple_quanlity)
    if (ok) then
        if (msg) then
            return tonumber(msg)
        else
            return 0
        end
    end

    utils.log_error("get device 3tuple quanlity failed" .. msg)
    utils.error_exit(500, "get device tuple data failed")
end


local function set_new_dev_config(drv_config, driver_name, pk, dn, config)
    local dev_config = {}
    dev_config["ProductKey"] = pk
    dev_config["DeviceName"] = dn
    dev_config["DeviceConfig"] = config
    table.insert(drv_config["DeviceList"], dev_config)
    utils.set_driver_config(driver_name, json.encode(drv_config))
end


function utils_device.set_device_config(driver_name, pk, dn, config)
    local drv_config_raw = utils.get_driver_config(driver_name)
    if (not drv_config_raw) then
        set_new_dev_config({DeviceList = {}}, driver_name, pk, dn, config)
        return true
    end

    local ok, msg = pcall(json.decode, drv_config_raw)
    if (ok) then
        local drv_config = msg
        if (drv_config["DeviceList"]) then
            for n, dev_config in ipairs(drv_config["DeviceList"]) do
                if (pk == dev_config["ProductKey"] and dn == dev_config["DeviceName"]) then
                    -- update device configuration
                    drv_config["DeviceList"][n]["DeviceConfig"] = config
                    utils.set_driver_config(driver_name, json.encode(drv_config))
                    return true
                end
            end
        else
            drv_config["DeviceList"] = {}
        end

        -- create new device configuration
        set_new_dev_config(drv_config, driver_name, pk, dn, config)
        return true
    end
    
    utils.log_error("invalid driver config" .. msg)
    utils.error_exit(500, "set config failed")
end


function utils_device.get_device_config(driver_name, pk, dn)
    local drv_config_raw = utils.get_driver_config(driver_name)    
    if (not drv_config_raw) then
        return {}
    end

    local ok, msg = pcall(json.decode, drv_config_raw)
    if (ok) then
        local drv_config = msg
        if (drv_config["deviceList"]) then
            for n, dev_config in ipairs(drv_config["deviceList"]) do
                if (pk == dev_config["productKey"] and dn == dev_config["deviceName"]) then
                    return dev_config["deviceConfig"]
                end
            end
        end
        return {}
    end
    
    utils.log_error("invalid driver config" .. msg)
    utils.error_exit(500, "get config failed")
end


return utils_device
