
local json = require("json")
local http = require("resty.http")
local utils = require("utils")
local const = require("const")

local request_method = ngx.var.request_method

utils.check_login()
utils.check_permission("ComputeManage") 

if (request_method == "GET") then

    local args = ngx.req.get_uri_args()
    if (nil == args) then
        utils.error_exit(400, "invalid parameter")
    end

    page_size = tonumber(args["PageSize"])
    current_page = tonumber(args["CurrentPage"])

    utils.check_current_page(current_page)
    utils.check_page_size(page_size)

    local url_joblist = const.kapacitor.base_url .. '/v1/nova/jobs'
   
    local httpc = http.new()
    local res, err = httpc:request_uri(url_joblist, {method = "GET"})
    if not res then
        utils.log_error("get kapacitor job list failed" .. err)
        utils.error_exit(500, "get kapacitor job list failed")
    end

    print(res.body)

    ok, job_list = pcall(json.decode, res.body)
    if (not ok) then
        utils.log_error("data from kapacitor is invalid" .. res.body)
        utils.error_exit(500, "data from kapacitor is invalid")
    end

    local res = {}
    res["PageSize"] = page_size
    res["CurrentPage"] = current_page
    res["List"] = {}

    local total = #job_list
    table.sort(job_list, function(e1, e2) return e1 < e2 end)
    local min, max = utils.get_item_index(total, page_size, current_page)
    for i = min, max do
        local job = {}
        job["JobId"] = job_list[1]
        job["Name"] = job_list[1]
        job["LoadTime"] = tostring(os.time())
        table.insert(res["List"], job)
    end 
   
    res_raw = json.encode(res)
    utils.error_exit(200, res_raw)
end

utils.error_exit(400, "unsupported request method")
------ EOF ------
