local json = require("json")
local utils = require("utils")
local const = require("const")
local pl_pretty = require("pl.pretty")
local resty_string = require("resty.string")
local pl_path = require("pl.path")
local pl_util = require("pl.utils")

local request_method = ngx.var.request_method

utils.check_login()
_, login_token = utils.check_permission("DriverManage")

local function check_param(name, value)
    
    if name == nil or type(name) ~= "string" then
        local ret = name.." must not be null"
        utils.error_exit(400, ret)
    end
end

if "POST" == request_method then
    local utils_header = require("utils_header")
    local params = utils_header.check_public_boby_data()
    local driverName = params["DriverName"]
    
    check_param("driverName", driverName)

    if driverName == "opcua" or driverName == "Modbus" or driverName == "websocket" then
        utils.error_exit(400, "system driver should not be removed")
    end

    local msg = utils.get_driver_config(driverName)
    if msg == nil or type(msg) ~= "string" then
        utils.error_exit(400, "driver config is not exist")
    end

    local msg_json = json.decode(msg)
    local file_path = msg_json["driverConfig"]["DriverZipPackagePath"]
    
    if file_path == nil then
        utils.error_exit(400, "driver config is not exist")
    end

    if not pl_path.isfile(file_path) then 
        utils.error_exit(400, "file is not exist.")
    end

    local cmd = "killall "..driverName.." ; rm -f "..file_path
    local ret, code = pl_util.execute(cmd)
    
    if not ret then
        utils.error_exit(500, json.encode("failed to run driver: "..code))
    else
        utils.set_driver_config(driverName, "")
        --unregister driver from dimu
        local dimu_cmd = string.format(const.dimu.method.fmt_unregisterDriver, driverName)
        local ret_code, ret_msg = utils.dbus_call_edge(const.dimu.wkn,
                                                        const.dimu.obj,
                                                        const.dimu.inf,
                                                        const.dimu.method.unregisterDriver,
                                                        dimu_cmd)

        if ret_code == "ok" and ret_msg[1] ~= nil then
            utils.error_exit(200, json.encode("file removed cleanly")) 
        else
            utils.error_exit(200, json.encode("file removed, but driver list still exist.")) 
        end
    end 
end
utils.error_exit(400, "unsupported request method")

